MODULE advect_mod
  USE icosa
  IMPLICIT NONE 

  PRIVATE
  
  PUBLIC :: init_advect, compute_backward_traj, compute_gradq3d, compute_advect_horiz

CONTAINS

  !==========================================================================

  SUBROUTINE init_advect(normal,tangent,sqrt_leng)
    IMPLICIT NONE
    REAL(rstd),INTENT(OUT) :: normal(3*iim*jjm,3)
    REAL(rstd),INTENT(OUT) :: tangent(3*iim*jjm,3)
    REAL(rstd),INTENT(OUT) :: sqrt_leng(iim*jjm)

    INTEGER :: ij

!DIR$ SIMD
      DO ij=ij_begin,ij_end

          CALL cross_product2(xyz_v(ij+z_rdown,:),xyz_v(ij+z_rup,:),normal(ij+u_right,:))
          normal(ij+u_right,:)=normal(ij+u_right,:)/sqrt(sum(normal(ij+u_right,:)**2)+1e-50)*ne(ij,right)

          CALL cross_product2(xyz_v(ij+z_up,:),xyz_v(ij+z_lup,:),normal(ij+u_lup,:))
          normal(ij+u_lup,:)=normal(ij+u_lup,:)/sqrt(sum(normal(ij+u_lup,:)**2)+1e-50)*ne(ij,lup)

          CALL cross_product2(xyz_v(ij+z_ldown,:),xyz_v(ij+z_down,:),normal(ij+u_ldown,:))	 
          normal(ij+u_ldown,:)=normal(ij+u_ldown,:)/sqrt(sum(normal(ij+u_ldown,:)**2)+1e-50)*ne(ij,ldown)

          tangent(ij+u_right,:)=xyz_v(ij+z_rup,:)-xyz_v(ij+z_rdown,:) 
          tangent(ij+u_right,:)=tangent(ij+u_right,:)/sqrt(sum(tangent(ij+u_right,:)**2)+1e-50)

          tangent(ij+u_lup,:)=xyz_v(ij+z_lup,:)-xyz_v(ij+z_up,:) 
          tangent(ij+u_lup,:)=tangent(ij+u_lup,:)/sqrt(sum(tangent(ij+u_lup,:)**2)+1e-50)

          tangent(ij+u_ldown,:)=xyz_v(ij+z_down,:)-xyz_v(ij+z_ldown,:)
          tangent(ij+u_ldown,:)=tangent(ij+u_ldown,:)/sqrt(sum(tangent(ij+u_ldown,:)**2)+1e-50)

          sqrt_leng(ij) = sqrt(max(sum((xyz_v(ij+z_up,:) - xyz_i(ij,:))**2),sum((xyz_v(ij+z_down,:) - xyz_i(ij,:))**2), &
                                   sum((xyz_v(ij+z_rup,:) - xyz_i(ij,:))**2),sum((xyz_v(ij+z_rdown,:) - xyz_i(ij,:))**2),  &
                                   sum((xyz_v(ij+z_lup,:) - xyz_i(ij,:))**2),sum((xyz_v(ij+z_ldown,:) - xyz_i(ij,:))**2)) )
    ENDDO

  END SUBROUTINE init_advect

  !=======================================================================================

  SUBROUTINE compute_gradq3d(qi_in,sqrt_leng_in,gradq3d_out,xyz_i,xyz_v)
    USE trace
    USE omp_para
    IMPLICIT NONE
    REAL(rstd),INTENT(IN)  :: qi_in(iim*jjm,llm)
    REAL(rstd),INTENT(IN)  :: sqrt_leng_in(iim*jjm)
    REAL(rstd),INTENT(IN)  :: xyz_i(iim*jjm,3)
    REAL(rstd),INTENT(IN)  :: xyz_v(2*iim*jjm,3)
    REAL(rstd),INTENT(OUT) :: gradq3d_out(iim*jjm,llm,3) 
    REAL(rstd) :: maxq,minq,minq_c,maxq_c 
    REAL(rstd) :: alphamx,alphami,alpha	,maggrd
    REAL(rstd) :: leng1,leng2
    REAL(rstd) :: arr(2*iim*jjm)
    REAL(rstd) :: ar(iim*jjm)
    REAL(rstd) :: gradtri(2*iim*jjm,llm,3) 
    INTEGER :: ij,k,ind,l
    REAL(rstd)  :: qi(iim*jjm,llm)
    REAL(rstd)  :: sqrt_leng(iim*jjm)
    REAL(rstd) :: gradq3d(iim*jjm,llm,3) 
    REAL(rstd) :: detx,dety,detz,det
    REAL(rstd) :: A(3,3), a11,a12,a13,a21,a22,a23,a31,a32,a33
    REAL(rstd) :: x1,x2,x3
    REAL(rstd) :: dq(3)

    qi=qi_in
    sqrt_leng=sqrt_leng_in
    
    CALL trace_start("compute_gradq3d1")

    ! TODO : precompute ar, drop arr as output argument of gradq ?

    !========================================================================================== GRADIENT 
    ! Compute gradient at triangles solving a linear system
    ! arr = area of triangle joining centroids of hexagons
!     DO l = ll_begin,ll_end 
!!DIR$ SIMD
!      DO ij=ij_begin_ext,ij_end_ext
!!             CALL gradq(ij,l,ij+t_rup,ij+t_lup,ij+z_up,qi,gradtri(ij+z_up,l,:),arr(ij+z_up))
!!             CALL gradq(ij,l,ij+t_ldown,ij+t_rdown,ij+z_down,qi,gradtri(ij+z_down,l,:),arr(ij+z_down))
!             CALL gradq(ij,l,ij+t_rup,ij+t_lup,ij+z_up,qi,gradtri(ij+z_up,l,1),gradtri(ij+z_up,l,2),gradtri(ij+z_up,l,3),arr(ij+z_up))
!             CALL gradq(ij,l,ij+t_ldown,ij+t_rdown,ij+z_down,qi,gradtri(ij+z_down,l,1),gradtri(ij+z_down,l,2),gradtri(ij+z_down,l,3),arr(ij+z_down))
!       END DO
!    END DO

     DO l = ll_begin,ll_end 
!DIR$ SIMD
      DO ij=ij_begin_ext,ij_end_ext
!       CALL gradq(ij,l,ij+t_rup,ij+t_lup,ij+z_up,qi,gradtri(ij+z_up,l,1),gradtri(ij+z_up,l,2),gradtri(ij+z_up,l,3),arr(ij+z_up))

        
        A(1,1)=xyz_i(ij+t_rup,1)-xyz_i(ij,1);  A(1,2)=xyz_i(ij+t_rup,2)-xyz_i(ij,2); A(1,3)=xyz_i(ij+t_rup,3)-xyz_i(ij,3) 
        A(2,1)=xyz_i(ij+t_lup,1)-xyz_i(ij,1);  A(2,2)=xyz_i(ij+t_lup,2)-xyz_i(ij,2); A(2,3)=xyz_i(ij+t_lup,3)-xyz_i(ij,3) 
        A(3,1)=xyz_v(ij+z_up,1);               A(3,2)= xyz_v(ij+z_up,2);             A(3,3)=xyz_v(ij+z_up,3)
    
        dq(1) = qi(ij+t_rup,l)-qi(ij,l)
        dq(2) = qi(ij+t_lup,l)-qi(ij,l)
        dq(3) = 0.0


!        CALL determinant(A(1,1),A(2,1),A(3,1),A(1,2),A(2,2),A(3,2),A(1,3),A(2,3),A(3,3),det)

         a11=A(1,1) ; a12=A(2,1) ; a13=A(3,1)
         a21=A(1,2) ; a22=A(2,2) ; a23=A(3,2)
         a31=A(1,3) ; a32=A(2,3) ; a33=A(3,3)

         x1 =  a11 * (a22 * a33 - a23 * a32)
         x2 =  a12 * (a21 * a33 - a23 * a31)
         x3 =  a13 * (a21 * a32 - a22 * a31)
         det =  x1 - x2 + x3
                 
!        CALL determinant(dq(1),dq(2),dq(3),A(1,2),A(2,2),A(3,2),A(1,3),A(2,3),A(3,3),detx)

         a11=dq(1)  ; a12=dq(2)  ; a13=dq(3)
         a21=A(1,2) ; a22=A(2,2) ; a23=A(3,2)
         a31=A(1,3) ; a32=A(2,3) ; a33=A(3,3)

         x1 =  a11 * (a22 * a33 - a23 * a32)
         x2 =  a12 * (a21 * a33 - a23 * a31)
         x3 =  a13 * (a21 * a32 - a22 * a31)
         detx =  x1 - x2 + x3
        
!        CALL determinant(A(1,1),A(2,1),A(3,1),dq(1),dq(2),dq(3),A(1,3),A(2,3),A(3,3),dety)

         a11=A(1,1) ; a12=A(2,1) ; a13=A(3,1)
         a21=dq(1)  ; a22=dq(2)  ; a23=dq(3)
         a31=A(1,3) ; a32=A(2,3) ; a33=A(3,3)

         x1 =  a11 * (a22 * a33 - a23 * a32)
         x2 =  a12 * (a21 * a33 - a23 * a31)
         x3 =  a13 * (a21 * a32 - a22 * a31)
         dety =  x1 - x2 + x3

!        CALL determinant(A(1,1),A(2,1),A(3,1),A(1,2),A(2,2),A(3,2),dq(1),dq(2),dq(3),detz)

         a11=A(1,1) ; a12=A(2,1) ; a13=A(3,1)
         a21=A(1,2) ; a22=A(2,2) ; a23=A(3,2)
         a31=dq(1)  ; a32=dq(2)  ; a33=dq(3)

         x1 =  a11 * (a22 * a33 - a23 * a32)
         x2 =  a12 * (a21 * a33 - a23 * a31)
         x3 =  a13 * (a21 * a32 - a22 * a31)
         detz =  x1 - x2 + x3

        gradtri(ij+z_up,l,1) = detx
        gradtri(ij+z_up,l,2) = dety
        gradtri(ij+z_up,l,3) = detz
        arr(ij+z_up) = det
        
      ENDDO
      
      DO ij=ij_begin_ext,ij_end_ext


!        CALL gradq(ij,l,ij+t_ldown,ij+t_rdown,ij+z_down,qi,gradtri(ij+z_down,l,1),gradtri(ij+z_down,l,2),gradtri(ij+z_down,l,3),arr(ij+z_down))

        A(1,1)=xyz_i(ij+t_ldown,1)-xyz_i(ij,1);  A(1,2)=xyz_i(ij+t_ldown,2)-xyz_i(ij,2); A(1,3)=xyz_i(ij+t_ldown,3)-xyz_i(ij,3) 
        A(2,1)=xyz_i(ij+t_rdown,1)-xyz_i(ij,1);  A(2,2)=xyz_i(ij+t_rdown,2)-xyz_i(ij,2); A(2,3)=xyz_i(ij+t_rdown,3)-xyz_i(ij,3) 
        A(3,1)=xyz_v(ij+z_down,1);               A(3,2)= xyz_v(ij+z_down,2);             A(3,3)=xyz_v(ij+z_down,3)
 
        dq(1) = qi(ij+t_ldown,l)-qi(ij,l)
        dq(2) = qi(ij+t_rdown,l)-qi(ij,l)
        dq(3) = 0.0


!        CALL determinant(A(1,1),A(2,1),A(3,1),A(1,2),A(2,2),A(3,2),A(1,3),A(2,3),A(3,3),det)

         a11=A(1,1) ; a12=A(2,1) ; a13=A(3,1)
         a21=A(1,2) ; a22=A(2,2) ; a23=A(3,2)
         a31=A(1,3) ; a32=A(2,3) ; a33=A(3,3)

         x1 =  a11 * (a22 * a33 - a23 * a32)
         x2 =  a12 * (a21 * a33 - a23 * a31)
         x3 =  a13 * (a21 * a32 - a22 * a31)
         det =  x1 - x2 + x3
                 
!        CALL determinant(dq(1),dq(2),dq(3),A(1,2),A(2,2),A(3,2),A(1,3),A(2,3),A(3,3),detx)

         a11=dq(1)  ; a12=dq(2)  ; a13=dq(3)
         a21=A(1,2) ; a22=A(2,2) ; a23=A(3,2)
         a31=A(1,3) ; a32=A(2,3) ; a33=A(3,3)

         x1 =  a11 * (a22 * a33 - a23 * a32)
         x2 =  a12 * (a21 * a33 - a23 * a31)
         x3 =  a13 * (a21 * a32 - a22 * a31)
         detx =  x1 - x2 + x3
        
!        CALL determinant(A(1,1),A(2,1),A(3,1),dq(1),dq(2),dq(3),A(1,3),A(2,3),A(3,3),dety)

         a11=A(1,1) ; a12=A(2,1) ; a13=A(3,1)
         a21=dq(1)  ; a22=dq(2)  ; a23=dq(3)
         a31=A(1,3) ; a32=A(2,3) ; a33=A(3,3)

         x1 =  a11 * (a22 * a33 - a23 * a32)
         x2 =  a12 * (a21 * a33 - a23 * a31)
         x3 =  a13 * (a21 * a32 - a22 * a31)
         dety =  x1 - x2 + x3

!        CALL determinant(A(1,1),A(2,1),A(3,1),A(1,2),A(2,2),A(3,2),dq(1),dq(2),dq(3),detz)

         a11=A(1,1) ; a12=A(2,1) ; a13=A(3,1)
         a21=A(1,2) ; a22=A(2,2) ; a23=A(3,2)
         a31=dq(1)  ; a32=dq(2)  ; a33=dq(3)

         x1 =  a11 * (a22 * a33 - a23 * a32)
         x2 =  a12 * (a21 * a33 - a23 * a31)
         x3 =  a13 * (a21 * a32 - a22 * a31)
         detz =  x1 - x2 + x3

         gradtri(ij+z_down,l,1) = detx
         gradtri(ij+z_down,l,2) = dety
         gradtri(ij+z_down,l,3) = detz
         arr(ij+z_down) = det

      END DO
    END DO

!DIR$ SIMD
    DO ij=ij_begin,ij_end
       ar(ij) = arr(ij+z_up)+arr(ij+z_lup)+arr(ij+z_ldown)+arr(ij+z_down)+arr(ij+z_rdown)+arr(ij+z_rup)+1.e-50
    ENDDO
    CALL trace_end("compute_gradq3d1")
    CALL trace_start2("compute_gradq3d2")
      
    DO k=1,3
      DO l =ll_begin,ll_end
!DIR$ SIMD
      DO ij=ij_begin,ij_end
              gradq3d(ij,l,k) = ( gradtri(ij+z_up,l,k) + gradtri(ij+z_down,l,k)   +   &
                                 gradtri(ij+z_rup,l,k) +  gradtri(ij+z_ldown,l,k)   +   & 
                                 gradtri(ij+z_lup,l,k)+    gradtri(ij+z_rdown,l,k) ) / ar(ij)  
        END DO
      END DO
    ENDDO
    CALL trace_end2("compute_gradq3d2")
    CALL trace_start("compute_gradq3d3")

    !============================================================================================= LIMITING 
    DO l =ll_begin,ll_end
!DIR$ SIMD
      DO ij=ij_begin,ij_end
!             maggrd =  dot_product(gradq3d(ij,l,:),gradq3d(ij,l,:))
             maggrd = gradq3d(ij,l,1)*gradq3d(ij,l,1) + gradq3d(ij,l,2)*gradq3d(ij,l,2) + gradq3d(ij,l,3)*gradq3d(ij,l,3) 
             maggrd = sqrt(maggrd) 
             maxq_c = qi(ij,l) + maggrd*sqrt_leng(ij)
             minq_c = qi(ij,l) - maggrd*sqrt_leng(ij)
             maxq = max(qi(ij,l),qi(ij+t_right,l),qi(ij+t_lup,l),qi(ij+t_rup,l),qi(ij+t_left,l), &
                  qi(ij+t_rdown,l),qi(ij+t_ldown,l))
             minq = min(qi(ij,l),qi(ij+t_right,l),qi(ij+t_lup,l),qi(ij+t_rup,l),qi(ij+t_left,l), &
                  qi(ij+t_rdown,l),qi(ij+t_ldown,l))
             alphamx = (maxq - qi(ij,l)) ; alphamx = alphamx/(maxq_c - qi(ij,l) )
             alphamx = max(alphamx,0.0)
             alphami = (minq - qi(ij,l)); alphami = alphami/(minq_c - qi(ij,l))
             alphami = max(alphami,0.0) 
             alpha   = min(alphamx,alphami,1.0)
!             gradq3d(ij,l,:) = alpha*gradq3d(ij,l,:)
             gradq3d(ij,l,1) = alpha*gradq3d(ij,l,1)
             gradq3d(ij,l,2) = alpha*gradq3d(ij,l,2)
             gradq3d(ij,l,3) = alpha*gradq3d(ij,l,3)
       END DO
    END DO

  CALL trace_end("compute_gradq3d3")
  
  
  DO k=1,3
    DO l = ll_begin,ll_end
      DO ij=ij_begin,ij_end 
        gradq3d_out(ij,l,k)=gradq3d(ij,l,k)
      ENDDO
    ENDDO
  ENDDO
  
  CONTAINS

    SUBROUTINE gradq(n0,l,n1,n2,n3,q,dq1,dq2,dq3,det)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: n0,l,n1,n2,n3
    REAL(rstd), INTENT(IN)     :: q(iim*jjm,llm)
!    REAL(rstd), INTENT(OUT)    :: dq(3), det
    REAL(rstd), INTENT(OUT)    :: dq1,dq2,dq3,det
    REAL(rstd)    :: dq(3)
    
    REAL(rstd)  ::detx,dety,detz
    INTEGER    :: info
    INTEGER    :: IPIV(3)

    REAL(rstd) :: A(3,3)
    REAL(rstd) :: B(3)
    
    ! TODO : replace A by A1,A2,A3
    A(1,1)=xyz_i(n1,1)-xyz_i(n0,1);  A(1,2)=xyz_i(n1,2)-xyz_i(n0,2); A(1,3)=xyz_i(n1,3)-xyz_i(n0,3) 
    A(2,1)=xyz_i(n2,1)-xyz_i(n0,1);  A(2,2)=xyz_i(n2,2)-xyz_i(n0,2); A(2,3)=xyz_i(n2,3)-xyz_i(n0,3) 
    A(3,1)=xyz_v(n3,1);              A(3,2)= xyz_v(n3,2);            A(3,3)=xyz_v(n3,3)

    dq(1) = q(n1,l)-q(n0,l)
    dq(2) = q(n2,l)-q(n0,l)
    dq(3) = 0.0

    !    CALL DGESV(3,1,A,3,IPIV,dq(:),3,info)
!    CALL determinant(A(:,1),A(:,2),A(:,3),det)
!    CALL determinant(dq,A(:,2),A(:,3),detx)
!    CALL determinant(A(:,1),dq,A(:,3),dety)
!    CALL determinant(A(:,1),A(:,2),dq,detz)
!    dq(1) = detx
!    dq(2) = dety
!    dq(3) = detz 

    CALL determinant(A(1,1),A(2,1),A(3,1),A(1,2),A(2,2),A(3,2),A(1,3),A(2,3),A(3,3),det)
    CALL determinant(dq(1),dq(2),dq(3),A(1,2),A(2,2),A(3,2),A(1,3),A(2,3),A(3,3),dq1)
    CALL determinant(A(1,1),A(2,1),A(3,1),dq(1),dq(2),dq(3),A(1,3),A(2,3),A(3,3),dq2)
    CALL determinant(A(1,1),A(2,1),A(3,1),A(1,2),A(2,2),A(3,2),dq(1),dq(2),dq(3),dq3)

  END SUBROUTINE gradq

  !==========================================================================
!  PURE SUBROUTINE determinant(a1,a2,a3,det)
!    IMPLICIT NONE
!    REAL(rstd), DIMENSION(3), INTENT(IN) :: a1,a2,a3
!    REAL(rstd), INTENT(OUT) :: det
!    REAL(rstd) ::  x1,x2,x3
!    x1 =  a1(1) * (a2(2) * a3(3) - a2(3) * a3(2))
!    x2 =  a1(2) * (a2(1) * a3(3) - a2(3) * a3(1))
!    x3 =  a1(3) * (a2(1) * a3(2) - a2(2) * a3(1))
!    det =  x1 - x2 + x3
!  END SUBROUTINE determinant

   SUBROUTINE determinant(a11,a12,a13,a21,a22,a23,a31,a32,a33,det)
    IMPLICIT NONE
    REAL(rstd), INTENT(IN) :: a11,a12,a13,a21,a22,a23,a31,a32,a33
    REAL(rstd), INTENT(OUT) :: det
    REAL(rstd) ::  x1,x2,x3
    x1 =  a11 * (a22 * a33 - a23 * a32)
    x2 =  a12 * (a21 * a33 - a23 * a31)
    x3 =  a13 * (a21 * a32 - a22 * a31)
    det =  x1 - x2 + x3
  END SUBROUTINE determinant


  END SUBROUTINE compute_gradq3d

  ! Backward trajectories, for use with Miura approach
  SUBROUTINE compute_backward_traj(normal,tangent,ue,tau, cc)
    USE trace
    USE omp_para
    IMPLICIT NONE
    REAL(rstd),INTENT(IN)    :: normal(3*iim*jjm,3)
    REAL(rstd),INTENT(IN)    :: tangent(3*iim*jjm,3)
    REAL(rstd),INTENT(IN)    :: ue(iim*3*jjm,llm)
    REAL(rstd),INTENT(OUT)   :: cc(3*iim*jjm,llm,3) ! start of backward trajectory
    REAL(rstd),INTENT(IN)    :: tau

    REAL(rstd) :: v_e(3), up_e, qe, ed(3)    
    INTEGER :: ij,l

    CALL trace_start("compute_backward_traj")

    ! TODO : compute normal displacement ue*tau as hfluxt / mass(upwind) then reconstruct tangential displacement
    
    ! reconstruct tangential wind then 3D wind at edge then cc = edge midpoint - u*tau
    DO l = ll_begin,ll_end
!DIR$ SIMD
      DO ij=ij_begin,ij_end
             up_e =1/de(ij+u_right)*(                                                   &
                  wee(ij+u_right,1,1)*le(ij+u_rup)*ue(ij+u_rup,l)+                        &
                  wee(ij+u_right,2,1)*le(ij+u_lup)*ue(ij+u_lup,l)+                        &
                  wee(ij+u_right,3,1)*le(ij+u_left)*ue(ij+u_left,l)+                      &
                  wee(ij+u_right,4,1)*le(ij+u_ldown)*ue(ij+u_ldown,l)+                    &
                  wee(ij+u_right,5,1)*le(ij+u_rdown)*ue(ij+u_rdown,l)+                    & 
                  wee(ij+u_right,1,2)*le(ij+t_right+u_ldown)*ue(ij+t_right+u_ldown,l)+    &
                  wee(ij+u_right,2,2)*le(ij+t_right+u_rdown)*ue(ij+t_right+u_rdown,l)+    &
                  wee(ij+u_right,3,2)*le(ij+t_right+u_right)*ue(ij+t_right+u_right,l)+    &
                  wee(ij+u_right,4,2)*le(ij+t_right+u_rup)*ue(ij+t_right+u_rup,l)+        &
                  wee(ij+u_right,5,2)*le(ij+t_right+u_lup)*ue(ij+t_right+u_lup,l)         &
                  )
             v_e = ue(ij+u_right,l)*normal(ij+u_right,:) + up_e*tangent(ij+u_right,:)
             cc(ij+u_right,l,:) = xyz_e(ij+u_right,:) - v_e*tau

             up_e=1/de(ij+u_lup)*(                                                      &
                  wee(ij+u_lup,1,1)*le(ij+u_left)*ue(ij+u_left,l)+                        &
                  wee(ij+u_lup,2,1)*le(ij+u_ldown)*ue(ij+u_ldown,l)+                      &
                  wee(ij+u_lup,3,1)*le(ij+u_rdown)*ue(ij+u_rdown,l)+                      &
                  wee(ij+u_lup,4,1)*le(ij+u_right)*ue(ij+u_right,l)+                      &
                  wee(ij+u_lup,5,1)*le(ij+u_rup)*ue(ij+u_rup,l)+                          & 
                  wee(ij+u_lup,1,2)*le(ij+t_lup+u_right)*ue(ij+t_lup+u_right,l)+          &
                  wee(ij+u_lup,2,2)*le(ij+t_lup+u_rup)*ue(ij+t_lup+u_rup,l)+              &
                  wee(ij+u_lup,3,2)*le(ij+t_lup+u_lup)*ue(ij+t_lup+u_lup,l)+              &
                  wee(ij+u_lup,4,2)*le(ij+t_lup+u_left)*ue(ij+t_lup+u_left,l)+            &
                  wee(ij+u_lup,5,2)*le(ij+t_lup+u_ldown)*ue(ij+t_lup+u_ldown,l)           &
                  )
             v_e = ue(ij+u_lup,l)*normal(ij+u_lup,:) + up_e*tangent(ij+u_lup,:)
             cc(ij+u_lup,l,:) = xyz_e(ij+u_lup,:) - v_e*tau
            

             up_e=1/de(ij+u_ldown)*(                                                    &
                  wee(ij+u_ldown,1,1)*le(ij+u_rdown)*ue(ij+u_rdown,l)+                    &
                  wee(ij+u_ldown,2,1)*le(ij+u_right)*ue(ij+u_right,l)+                    &
                  wee(ij+u_ldown,3,1)*le(ij+u_rup)*ue(ij+u_rup,l)+                        &
                  wee(ij+u_ldown,4,1)*le(ij+u_lup)*ue(ij+u_lup,l)+                        &
                  wee(ij+u_ldown,5,1)*le(ij+u_left)*ue(ij+u_left,l)+                      & 
                  wee(ij+u_ldown,1,2)*le(ij+t_ldown+u_lup)*ue(ij+t_ldown+u_lup,l)+        &
                  wee(ij+u_ldown,2,2)*le(ij+t_ldown+u_left)*ue(ij+t_ldown+u_left,l)+      &
                  wee(ij+u_ldown,3,2)*le(ij+t_ldown+u_ldown)*ue(ij+t_ldown+u_ldown,l)+    &
                  wee(ij+u_ldown,4,2)*le(ij+t_ldown+u_rdown)*ue(ij+t_ldown+u_rdown,l)+    &
                  wee(ij+u_ldown,5,2)*le(ij+t_ldown+u_right)*ue(ij+t_ldown+u_right,l)     &
                  )
             v_e = ue(ij+u_ldown,l)*normal(ij+u_ldown,:) + up_e*tangent(ij+u_ldown,:)
             cc(ij+u_ldown,l,:) = xyz_e(ij+u_ldown,:) - v_e*tau
       ENDDO
    END DO

    CALL trace_end("compute_backward_traj")

  END SUBROUTINE compute_backward_traj

  ! Horizontal transport (S. Dubey, T. Dubos)
  ! Slope-limited van Leer approach with hexagons
  SUBROUTINE compute_advect_horiz(update_mass,diagflux_on, hfluxt,cc,gradq3d, mass,qi,qfluxt)
    USE trace
    USE omp_para
    IMPLICIT NONE
    LOGICAL, INTENT(IN)       :: update_mass, diagflux_on
    REAL(rstd), INTENT(IN)    :: gradq3d(iim*jjm,llm,3) 
    REAL(rstd), INTENT(IN)    :: hfluxt(3*iim*jjm,llm) ! mass flux
    REAL(rstd), INTENT(IN)    :: cc(3*iim*jjm,llm,3)   ! barycenter of quadrilateral, where q is evaluated (1-point quadrature)
    REAL(rstd), INTENT(INOUT) :: mass(iim*jjm,llm)
    REAL(rstd), INTENT(INOUT) :: qi(iim*jjm,llm)
    REAL(rstd), INTENT(INOUT) :: qfluxt(3*iim*jjm,MERGE(llm,1,diagflux_on)) ! time-integrated tracer flux

    REAL(rstd) :: dq,dmass,qe,ed(3), newmass
    REAL(rstd) :: qflux(3*iim*jjm,llm)
    INTEGER :: ij,k,l

    CALL trace_start("compute_advect_horiz")
#include "../kernels/advect_horiz.k90"
    CALL trace_end("compute_advect_horiz")
  END SUBROUTINE compute_advect_horiz


END MODULE advect_mod
